/*----------------------------------------------------------------------

  SQL Functions()
  
- isSQL() -- really SQL? must find a keyword select, distinct, etc 
- rules
	- '\n' with each WHERE, FROM, ORDER by....and color blue
	-  color code data too...

http://www.w3schools.com/SQl/sql_quickref.asp

*-------------------------------------------------------------------------*/

#include "LogFile.h"

/* Safeties */
#define MAX_DATA 200	 // Runaway data detection.
#define MAX_SQL_RUN 5000 // Runaway SQL detection: don't stay in this module longer this many chars.


struct tagKeyWords
{
	char *keyWord;
	int newLine;
	long freq;
};  

static struct tagKeyWords kw[44] = 
{
/* These are arranged by most common 1st, for faster search
** They are also broken into 2 groups ...unique to SQL 1st,
** and common to English 2nd.  This helps enable the isSQL() function */
{	"SELECT", 	1,	241133	},  //* sysdat =   241133/7=
{	"WHERE"	, 	1,	216098	},
{	"NULL"	, 	0,	154266	},
{	"DISTINCT",	0,	81590	},
{	"INSERT", 	1,	34646	},
{	"INTO"	, 	0,	34646	},
{	"VALUES", 	1,	34644	},
{	"SYSDATE", 	0,	34447	},
{	"DUAL"	, 	0,	33027	},
{	"COUNT"	, 	0,	26609	},
{	"SET"	, 	1,	7277	},
{	"UPDATE", 	1,	7249	},
{	"ORDER"	, 	1,	6544	},
{	"LIKE"	, 	0,	1832	},
{	"DELETE", 	1,	716	},
{	"JOIN"	, 	0,	632	},
{	"DECODE",	0,	0	},
/*  This divides common English and SQL ---*/
#define BEGIN_COMMON_ENGLISH_IDX 17
{	"AND"	, 	1,	339526	},
{	"FROM"	, 	1,	241849	}, /* Hate to leave 'FROM' down here */
{	"IS"	, 	0,	100457	},
{	"AS"	, 	0,	44166	},
{	"IN"	, 	0,	20463	},
{	"OR"	, 	1,	10277	},
{	"BY"	, 	0,	6544	},
{	"NOT"	, 	0,	721	},
{	"ON"	, 	1,	717	},
{	"LEFT"	, 	1,	632	},
/* These next are certainly more SQL than English, but not seen much, 
* so placed at end for run-time speed optimization*/
{	"CREATE", 	1,	0	},
{	"UPSERT", 	1,	0	},
{	"TABLE"	, 	0,	0	},
{	"MERGE"	, 	1,	0	},
{	"TRUNCATE",	1,	0	},
{	"BETWEEN", 	0,	0	},
{	"UNION"	, 	1,	0	},
{	"ALTER"	, 	1,	0	},
{	"ADD"	, 	1,	0	},
{	"DROP"	, 	1,	0	},
{	"GROUP"	, 	1,	0	},
{	"INNER"	, 	1,	0	},
{	"OUTER"	, 	1,	0	},
{	"RIGHT"	, 	1,	0	},
{	"FULL"	, 	1,	0	},
{	"HAVING", 	1,	0	},
{NULL, 0, 0 }
};


/* First letter of Key Word string...speed optimization */
#define MAX_KWLEN  8



/*----------------------------------------------------------------------

	isSQLalnum(int ch) 

  Legal identifier / name chars:   After the first character, 
  identifiers can be alphanumeric, and $#@_ 

*-------------------------------------------------------------------------*/
int isSQLalnum(int ch)
{
	/* This is much faster than the str search, also, it is more accurate than 
	calling isalnum (ch) ... which is flawed in MSVC++ 6.0 std library) */

	if(ch >= 'A' && ch <= 'Z')
		return 1;
	if(ch >= 'a' && ch <= 'z')
		return 1;
	if(ch >= '0' && ch <= '9')
		return 1;
	switch(ch)
	{
		case '$':
		case 'X':
		case '_':
		case '@':
			return 1;
		break;
		default:
			return 0;
		break;
	}
	return 0;
}


/*----------------------------------------------------------------------

	printSQLkwFreq()
	
	Print KeyWord frequency stats  to a file
	(Development use only)
	  
*-------------------------------------------------------------------------*/
void printSQLkwFreq(FILE *outfPtr)
{
int iii;
		/*{"SELECT",1, 0 },*/

	for(iii=0; kw[iii].keyWord; iii++) 
		fprintf(outfPtr,"{:*%s*:, :%d:,:%ld:},\r\n", kw[iii].keyWord,kw[iii].newLine, kw[iii].freq);
}


/*----------------------------------------------------------------------

	isKW() 
	-- is this a keyword?
	  
*-------------------------------------------------------------------------*/
int isKW(const char *inStr, int len , int *newline)
{
int iii, kwlen;

	for(iii=0; kw[iii].keyWord; iii++) 
	{
		kwlen = strlen(kw[iii].keyWord);
		if(len > kwlen &&  ! isSQLalnum( inStr[kwlen] ) )
		{
			if(!strnicmp(inStr, kw[iii].keyWord, min(len,kwlen)) )
			{
				if(*newline) /* if we already have a newline, don't tell tell caller to write one */
					*newline = FALSE;
				else
					*newline = kw[iii].newLine; /* if not newline, set the optional newline */
				/* update frequency */
				/* kw[iii].freq++;  Dev only*/
				return kwlen;
			}
		}
	}
	return 0;
}
/*----------------------------------------------------------------------

- isSQL() -- really SQL? must find a keyword select, etc
  
  Here's a concept for SQL detection algorythm:

  Possible SQL?  then...Scan until ';'
  rules 
  1. ending ';' must be withing XX chars (if < XX chars..!SQL
  2. in that space, must find 2 keywords
  3. keys words might be / should be seperated by non-keywords
  4. SQL stray/forbiden chars must not exist in buffer (this mite not be reliable rule either way)
  5. if 'UPDATE'...peculiar handling for common launch commands

Most common

SELECT DISTINCT FROM WHERE AND NULL

  Added a long value to the keyword structure to make a frequency histogram.
  Update in the isKW() function
  Sorted the table (static) by frequency, for faster matches

AND		141371
FROM	85461
SELECT	85295
WHERE	70335  <<< note that "where" does not always accompany "select from"
NULL	58156
IS		44812
DISTINCT	27784
AS		17469
DUAL	17310
INSERT	9110
INTO	9110
VALUES	9108
IN		8974
COUNT	6803
OR		5178
SET		2184
UPDATE	2156
ORDER	2101
BY		2101
LIKE	947
NOT		281
ON		228
JOIN	212
LEFT	212
DELETE	166
CREATE	0
UPSERT	0
TABLE	0
MERGE	0
TRUNCATE	0
BETWEEN	0
UNION	0
ALTER	0
ADD		0
DROP	0
GROUP	0
INNER	0
OUTER	0
RIGHT	0
FULL	0
HAVING	0

  Currently calling like this
  if(toupper(curCh)== 'S' && isSQL(&g_readBuff[g_CurOffs], 55 ))
-------------------------------------------------------------------------*/
enum {
	SQL = 1,
	SELECT,
	INSERT,
	UPDATE,
	_DELETE,
	CREATE
};
struct tagSQLfw
{
	char *kw;
	char *nextKW;
	int type;
};  

static struct tagSQLfw SQL_fw[7]={
{"SQL",	NULL,    SQL        },
{"SELECT", "from",	SELECT  },
{"INSERT", "into",	INSERT  },
{"UPDATE", "set",	UPDATE  },
{"DELETE", "from",	_DELETE },
{"CREATE", "table", CREATE  },
{NULL, NULL, 0}
};

int isSQL(const char *inStr, int lastCh)
{
int iii=0, type = 0, kwCount =0, strl =0, maxScan;
char *ptr=NULL;

	/* Hit the most common 1st ...new __strnistr(upper/lower, lower only, count)*/
	/*Remember 1st and last char (word seperators) must be space or special chars!!*/

	maxScan = __memnchr(inStr, ';', MIN_LOOK_AHEAD );
	if(maxScan < 15)
		return 0;

	switch(toupper(*inStr))
	{
		case 'S':
		case 'I':
		case 'U':
		case 'D':
		case 'C':
			iii=0;
			while(!kwCount && SQL_fw[iii].kw)
			{
				strl = strlen(SQL_fw[iii].kw );
				if(!isSQLalnum( inStr[strl] ) &&  !isSQLalnum( lastCh ) && 
					!strnicmp(inStr, SQL_fw[iii].kw, strl)   ) 
				{
					type = SQL_fw[iii].type;
					kwCount++;
					maxScan -= strl;
					inStr+=strl;

					if(type == SQL)
						break; /* N/A for nextKW likely check */

					/* Now, look for next likely key word */
					strl = strlen(  SQL_fw[iii].nextKW   );
					if((ptr= __strnistr(inStr, SQL_fw[iii].nextKW, maxScan))  && 
							!isSQLalnum( ptr[strl] ) && 
							!isSQLalnum( ptr[-1] ))
						return TRUE;  /* Have 2 keywords, bail...*/
					
					}
				iii++;
			}
		break;
		default :
		break;
	}
	if(maxScan < 12)
		return 0;

	if(!kwCount)
		return 0;

	/* We enter  this loop when the first word is some key word 
	** (or 'SQL'). Scan ahead looking for next keyword...  
	***/
	iii=0;
	while(kwCount < 2  && iii < BEGIN_COMMON_ENGLISH_IDX)
	{
	/* Check previous and end char are not identifier legal chars, as well as kw match */
		if((ptr= __strnistr(inStr, kw[iii++].keyWord, maxScan))  && 
				!isSQLalnum( ptr[strl] ) && 
				!isSQLalnum( ptr[-1] ))
					kwCount++;
	}
	return kwCount > 1;

}

static int newline =0;
static int chars_this_line =0;

/*----------------------------------------------------------------------


  isNewLineOpp()
	is this a newline opportunity? (our wordwrap helper)

*-------------------------------------------------------------------------*/
int isNewLineOpp(int ch)
{
	if(ch == '.' || isSQLalnum( ch ) )
		return FALSE;
	return TRUE;
}


/*----------------------------------------------------------------------


  NewLineStdChks()
	perform standard checks for OK to write newline.  
	Consolidate redundant code that looked like this:

    if(chars_this_line > 80  &&  isNewLineOpp(curCh) && nextCh != ';' )
	if(chars_this_line > 80  && !indata && nextCh != ';' && nextCh != '\n')
	if(chars_this_line > 80 &&  !newline  && !indata && nextCh != ';')

  Any parameter can be overridden to modify return
*-------------------------------------------------------------------------*/
int NewLineStdChks(int chars_this_line, int curCh, int nextCh, int indata)
{
int betterOppComing = 0;

	switch(nextCh)
	{
		case '\n':
		case ';':
		case ',':
			betterOppComing = 1;
		break;
		default:
			betterOppComing = 0;
		break;
	}
	if(chars_this_line > 80  && 
		!indata && 
		!betterOppComing && 
		isNewLineOpp(curCh) 
		)
		return TRUE;
	return FALSE;


}


/*----------------------------------------------------------------------


  wCR()

  local implementation of write carriage return

*-------------------------------------------------------------------------*/
static void wCR(FILE *outfPtr)
{
	newline = TRUE;
	chars_this_line =0;
	if(plain_text)
		fwrite("\r\n", 1, 2, outfPtr);
	else
		fwrite("<br>\r\n", 1, 6, outfPtr);
}

/*----------------------------------------------------------------------


  strFmtSQL()

	- rules
	- CR before certain keywords: SELECT, WHERE, FROM, ORDER by, etc
	- Color-code data and keywords Blue...else Black.

*-------------------------------------------------------------------------*/
int strFmtSQL(FILE *infPtr, FILE *outfPtr, int  *inCurCh, long dwChCount, long fLen)
{
int curCh, lastCh=0, nextCh, kwlen, bytesRead=1;
int done = FALSE, needNewLine;
long initialCount = dwChCount, inDataCount=0;
int indata = FALSE;
char buff[200];

	newline = TRUE;
	chars_this_line =0;
	needNewLine = indata = FALSE;

	if(plain_text)
		fwrite("\r\n\r\n\r\n", 1, 6, outfPtr);
	else
		fprintf(outfPtr,"%s<p>", SQL_KeyWord_Color? "<FONT COLOR=#000000>":"");

	curCh = *inCurCh;
	while(!done)
    {
		nextCh = g_readBuff[g_CurOffs+1];
		bytesRead=1;

		if(!(dwChCount % 100))
			SpinTheMessageLoop(NULL);
		if( gQuitMssg )
			break;

        switch(curCh)
        {
			case ';':
				if(indata)
				{
					/* This is a suspect case....what if a double set of '' 
					** erroneously arrived in the logfile and we're in runaway mode?
					** This is the most elegant place to catch this.  We can also catch
					** later at bottom loop.  Good to keep both in case we don't see 
					** ';' for a long time
					*/
					if(inDataCount >= MAX_DATA)
					{
						if(SQL_KeyWord_Color)
							fprintf(outfPtr,"<FONT COLOR=#000000>");
						indata = FALSE;
					}
				}
				if(!indata)
				{
					done = TRUE;
				}
				fputc(curCh,outfPtr);
			break;

            case '\'':
				if(!indata && nextCh != '\'') //don't chng color for empty ''
				{
					fprintf(outfPtr,"'%s", SQL_KeyWord_Color? "<FONT COLOR=#0011EE>":"");
				}
				else
				{
					if(indata && lastCh != '\'') //don't chng color for empty ''
						fprintf(outfPtr,"%s'", SQL_KeyWord_Color? "<FONT COLOR=#000000>":"");
					else
						fputc(curCh,outfPtr); /* Still have not printed... do it */ 
				}
				indata = !indata;
				newline = FALSE;
				inDataCount = 0;
            break;

            case '=':
				if(!indata && SQL_KeyWord_Color)
				{
					fprintf(outfPtr,"<FONT COLOR=#0011EE>=<FONT COLOR=#000000>");
				}
				else
					fputc(curCh,outfPtr);
				newline = FALSE;
            break;

            case 0x0A:  /* Toss LFs until i want (note no 0x0D case..stripped by fmtBuffer()*/
				if(NewLineStdChks(chars_this_line, curCh, nextCh, indata) )
				{
					wCR(outfPtr);
				}
				else /* Bug fix - skipping CRs can cause words to join */
				{
					fputc(' ',outfPtr); /* write a space */
				}
            break;

            case ' ':
				if(newline) /* ignore space at start of newline */
					break;
				fputc(curCh,outfPtr);
				newline = FALSE;

				if(NewLineStdChks(chars_this_line, curCh, nextCh, indata) )
				{
					wCR(outfPtr);
				}
			break;
			case '|':
				if(nextCh == '|' && !indata)
				{
					/* This is the concatenation symbol '||' */
					bytesRead = 2;
					if(SQL_KeyWord_Color)
						fprintf(outfPtr,"<FONT COLOR=#0011EE>||<FONT COLOR=#000000>");
					else
						fprintf(outfPtr,"||");
					break;
					if(NewLineStdChks(chars_this_line+2, curCh, nextCh, indata) )
						wCR(outfPtr); 
				}
				fputc(curCh,outfPtr);
				newline = FALSE;
			break;


			case EOF:
				done = TRUE;
			break;

            default:
				needNewLine = FALSE;
				kwlen = 0;
				if(!indata)
				{
					if(NewLineStdChks(chars_this_line, curCh, nextCh, indata) )
					{
						needNewLine = TRUE;
					}
					if( !isSQLalnum( lastCh ) )    /* See if we have a keyword */
					{
						kwlen = isKW(&g_readBuff[g_CurOffs], MAX_KWLEN+3, &newline);
						if(kwlen)
						{
						char *pNL ="<br>";

							if(plain_text)
								pNL ="\r\n";

							strncpy(buff, &g_readBuff[g_CurOffs], kwlen);
							buff[kwlen]='\0';
							if(SQL_KeyWord_Color)
								fprintf(outfPtr,"%s<FONT COLOR=#0011EE>%s<FONT COLOR=#000000>", newline? pNL:"", buff);
							else
								fprintf(outfPtr,"%s%s", newline?pNL:"", buff);
							if(newline)
							{
								chars_this_line=0;
								needNewLine = FALSE;
							}
							bytesRead = kwlen; 
						}
					}
				}
				if(!kwlen)
					fputc(curCh,outfPtr);
				
				newline = FALSE;
				
				if(needNewLine )
				{
					wCR(outfPtr);
				}

            break;
        }
		lastCh = curCh;

		if(!newline)
			chars_this_line += bytesRead;
		dwChCount += bytesRead;
		g_CurOffs += bytesRead;

		/*-------------------------------------------------
		Runaway ck: See if we've been in data for excessive time.
		200 chars is probably bigger than any field.
		This is necessary due to log errors that double the '' char
		data on/off toggle...we may no longer be in data...dangerous as
		it can also cause us to miss our ';' quit signal
		*----------------------------------------------------*/
		if(indata)
		{
			inDataCount += bytesRead;
			if(inDataCount > MAX_DATA+10)
			{
				if(SQL_KeyWord_Color)
					fprintf(outfPtr,"<FONT COLOR=#000000>");
				indata = FALSE;
				inDataCount = 0;
			}
		}
		else
			inDataCount = 0;

		if(dwChCount-initialCount > MAX_SQL_RUN )
		{
			/* This is the longest SQL statement known to humankind, or we are in runaway 
			** due to missing ';' stop sign.  ...lets quit SQL formatting...*/
			done = TRUE;
		}

		if(done || (dwChCount >= fLen)) /* we're done and we don't know it! */
            break;

		if(!LoadBuffer(infPtr))
			break;

		curCh = g_readBuff[g_CurOffs];
	}
	*inCurCh = curCh;
	
	if(plain_text)
		fwrite("\r\n", 1, 2, outfPtr);
	else
		fprintf(outfPtr,"%s</p>", SQL_KeyWord_Color? "<FONT COLOR=#000000>":"");
	return dwChCount-initialCount; /* return bytes read */

}