
#include "LogFile.h"


/*----------------------------------------------------------------------

  char *strnistr()

   returns pointer to first occurrance of substring in string
    NULL if not found

 params:

      str - string to search
      pattern - string to search for

 This optimized version is twice as fast as competitor... about 48
  bytes more expensive in code.

 Also discovered that tolower() is faster than toupper() 
----------------------------------------------------------------------*/
char *__strnistr(const char *str, const char *pattern, int len)
{
int p1stCh, sCh, sOffs =0, maxsOffs;
int found = 0, done = 0, patlen;
    
	if(!str || !pattern || !*pattern)	
		return NULL;

	patlen=strlen(pattern);

	/* Beyond this offset, no match can be made due to  
	** lack sufficient remaining search length. Safety + optimization*/
	maxsOffs = len-patlen;

	p1stCh = tolower(*pattern);

	if(len)
	{
		do {
			found = 0;
			// Speed optimize: scan string for a match on 1st character
			while(!done && !found)
			{
				sCh = tolower(str[sOffs++]);
				if(sOffs > maxsOffs  ||  !sCh)
					done = 1;
				/* we don't break -- we could have a 1-byte string match */
				if(sCh == p1stCh)
				{
					found = 1;
				}
			}

			if(found)
			{
			int f,l,result = 0, count = patlen-1;
			const char *string = &str[sOffs], *subStr = &pattern[1];

				if(count && *string && *subStr)
				{
					do{
						  f = tolower(*string++);
						  l = tolower(*subStr++);
						  if(! *subStr )  /* reached end of substring */
							  break;
					} while (--count && f && l && f == l);
					result = (f - l);
				}
				if(!result)
					return (char *)(&str[sOffs-1]);
			}
		}while(!done );
	}
    return(NULL);
}




/*------------------------------------------------------------

  strnchr()

  returns a pointer to the first occurrence of ch in string, 
  or NULL if c is not found.

----------------------------------------------------------------*/
char *__strnchr(  char *str, int ch, int len )
{
int iii=0;

	if(!str)
		return NULL;
	
	while (iii < len && str[iii] && str[iii] != ch )
	{
		iii++;
	}

	if(str[iii] == ch)
		return &str[iii];

	return NULL;
}


/*------------------------------------------------------------

  _memnchr()

  returns an integer offset to the first occurrence of ch in string, 
  
	returns -1 not found.

----------------------------------------------------------------*/
int __memnchr( const char *str, int ch, int len )
{
int iii=0, offs = -1;

	while (iii < len  )
	{
		if(str[iii] == ch )
		{
			offs = iii;
			break;
		}
		iii++;
	}
	return offs;
}







